
#include "esp_camera.h"
#include <TMCStepper.h>
#include <WiFi.h>
#include <ESPAsyncWebServer.h>
#include <ArduinoJson.h>
#include "esp_timer.h"
#include "img_converters.h"
#include <SPIFFS.h>
#include <FS.h>
#include "soc/soc.h"           // Disable brownout problems
#include "soc/rtc_cntl_reg.h"  // Disable brownout problems
#include <ESPmDNS.h>
#include <Update.h>

// Yes ESP32cam

#define U_PART U_SPIFFS

const char *SSID = "NetworkCongestion";
const char *PWD = "P@ssword";
const char *hostname = "BarnDoor";

#define SERIAL_PORT Serial2 // HardwareSerial port 
#define DRIVER_ADDRESS 0b00
#define ENABLE_PIN 14
#define RED_LED 33
// #define TOUCH_PIN 12

#define DIAG_PIN 13
#define INDEX_PIN 12

#define STALL_VALUE  10 // [-64..63]

#define RXD2 15
#define TXD2 2

#define FILE_PHOTO "/photo.jpg"

// using namespace TMC2209_n;

#define R_SENSE 0.11f // Match to your driver
                      // SilentStepStick series use 0.11
                      // UltiMachine Einsy and Archim2 boards use 0.2
                      // Panucatt BSD2660 uses 0.1
                      // Watterott TMC5160 uses 0.075


// OV2640 camera module pins (CAMERA_MODEL_AI_THINKER)
#define PWDN_GPIO_NUM     32
#define RESET_GPIO_NUM    -1
#define XCLK_GPIO_NUM      0
#define SIOD_GPIO_NUM     26
#define SIOC_GPIO_NUM     27

#define Y9_GPIO_NUM       35
#define Y8_GPIO_NUM       34
#define Y7_GPIO_NUM       39
#define Y6_GPIO_NUM       36
#define Y5_GPIO_NUM       21
#define Y4_GPIO_NUM       19
#define Y3_GPIO_NUM       18
#define Y2_GPIO_NUM        5
#define VSYNC_GPIO_NUM    25
#define HREF_GPIO_NUM     23
#define PCLK_GPIO_NUM     22

TMC2209Stepper driver(&SERIAL_PORT, R_SENSE, DRIVER_ADDRESS);
AsyncWebServer server(80);

int accel = 0;
int stall = STALL_VALUE;
bool shaft = false;
volatile int direction = 0;

bool hascamera = false;
bool takepicture = false;

unsigned long opentiming = 14500;
unsigned long closetiming = 14500;

volatile bool bumped = false;
// volatile bool touch1detected = false;

volatile long steps;
size_t content_len;

TaskHandle_t Motor;

void IRAM_ATTR bump(void* arg)
{
    bumped = true;
}

void IRAM_ATTR step(void* arg)
{
    steps = steps +1;
}

// void gotTouch1(){
//  touch1detected = true;
// }

// web page

const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML><html>
<head>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <style>
    body { text-align:center; }
    .vert { margin-bottom: 10%; }
    .hori{ margin-bottom: 0%; }
    .center {
       margin-left: auto;
       margin-right: auto;
      }
  </style>
</head>
<body>
  <div id="container">
    <h2>the Lab</h2>
    <p>
      <button onclick="OpenDoor();">Open Door</button>
      <button onclick="CloseDoor()">Close Door</button>
      <button onclick="location.reload();">REFRESH PAGE</button>
      <button onClick="UpdateFirmware()">Update Firmware</button>
    </p>
  </div>
  <div id="stats"></div>
  <div><img onclick="TakePic();" src="saved-photo" id="photo" width="70%"></div>
</body>
<script>

  function FillStats(arr) {
    var out = "<table class='center'>";
    var i;
    for(i = 0; i < arr.length; i++) {
        out += '<tr><td>' + arr[i].type + '</td><td>' + arr[i].value + '</td>' + '</tr>';
    }
    out += "</table";
    document.getElementById("stats").innerHTML = out;
  }

  function OpenDoor() {
    var xhr = new XMLHttpRequest();
    xhr.open('GET', "/Open", true);
    xhr.send();
  }

  function CloseDoor() {
    var xhr = new XMLHttpRequest();
    xhr.open('GET', "/Close", true);
    xhr.send();
  }

  function TakePic(){
    xhr = new XMLHttpRequest();
    xhr.onreadystatechange = function() {
        if (this.readyState == 4 && this.status == 200) {
            var myArr = JSON.parse(this.responseText);
            FillStats(myArr);
        }
    };
    xhr.open('GET', "/pic", true);
    xhr.send(); 
  }
  function UpdateFirmware(){
    document.getElementById("stats").innerHTML ="<form method='POST' action='/doUpdate' enctype='multipart/form-data'><input type='file' name='update'><input type='submit' value='Update'></form>"
  }
</script>
</html>)rawliteral";

StaticJsonDocument<250> jsonDocument;
char buffer[250];

void setupMDNS()
{
  /*use mdns for host name resolution*/
  if (!MDNS.begin(hostname)) { //http://esp32.local
    Serial.println("Error setting up MDNS responder!");
    while (1) {
      delay(1000);
    }
  }
  Serial.println("mDNS responder started");
}

void Mount_Storage()
{
  if (!SPIFFS.begin(true)) {
    Serial.println("An Error has occurred while mounting SPIFFS");
    ESP.restart();
  }
  else {
    delay(500);
    Serial.println("SPIFFS mounted successfully");
  }
}

void connectToWiFi() {
  Serial.print("Connecting to ");
  Serial.println(SSID);
  WiFi.mode(WIFI_STA);
  WiFi.config(INADDR_NONE, INADDR_NONE, INADDR_NONE, INADDR_NONE);
  WiFi.setHostname(hostname); //define hostname

  WiFi.begin(SSID, PWD);
  
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print(".");
    delay(500);
    // we can even make the ESP32 to sleep
  }
 
  Serial.print("Connected. IP: ");
  Serial.println(WiFi.localIP());
  Serial.print("RRSI: ");
  Serial.println(WiFi.RSSI());
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);
  setupMDNS();
}

// setup API resources
void setup_routing() {

  server.on("/Open", HTTP_GET, [](AsyncWebServerRequest * request) {
    OpenDoor(request);
  });

  server.on("/Close", HTTP_GET, [](AsyncWebServerRequest * request) {
    CloseDoor(request);
  });

  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html);
  });

  server.on("/pic", HTTP_GET, [](AsyncWebServerRequest *request){
    TakePic(request);
  });

  server.on("/saved-photo", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, FILE_PHOTO, "image/png");
  });
  
  server.on("/update", HTTP_GET, [](AsyncWebServerRequest *request){handleUpdate(request);});
  server.on("/doUpdate", HTTP_POST,
    [](AsyncWebServerRequest *request) {},
    [](AsyncWebServerRequest *request, const String& filename, size_t index, uint8_t *data,
                  size_t len, bool final) {handleDoUpdate(request, filename, index, data, len, final);}
  );

  // start server
  server.begin();
}

void setup_TMC2209()
{
  digitalWrite(ENABLE_PIN ,LOW);

  pinMode(DIAG_PIN ,INPUT);
   pinMode(INDEX_PIN, INPUT); //How far did it go?

  pinMode(RED_LED, OUTPUT);
  digitalWrite(RED_LED, HIGH); // off

  Serial2.begin(115200, SERIAL_8N1, RXD2, TXD2);     // INITIALIZE UART to TMC2208
  
  driver.begin();                // Initialize driver
                           
  driver.toff(5);                 // Enables driver in software

  driver.rms_current(800);       // Set motor RMS current
  Serial.print("current set to:");
  Serial.println(driver.rms_current());
 
  driver.microsteps(256); 
           
  driver.TCOOLTHRS(0xFFFFF); // 20bit max
  driver.TPWMTHRS(0); 
  driver.semin(5);
  driver.semax(2);
  driver.sedn(0b01);
  driver.SGTHRS(STALL_VALUE);
  driver.pwm_autoscale(true);   // Needed for stealthChop
  driver.en_spreadCycle(false);   // Toggle spreadCycle for smooth & silent operation
  driver.blank_time(24);
  driver.iholddelay(3); 

  //attachInterrupt(digitalPinToInterrupt(DIAG_PIN), bump, RISING);
  //attachInterrupt(digitalPinToInterrupt(INDEX_PIN), step, FALLING);

  //Some crap about the camera means more arcane looking code.
  gpio_install_isr_service(0);
  gpio_set_intr_type(GPIO_NUM_13, GPIO_INTR_POSEDGE); //DIAG_PIN
  gpio_isr_handler_add(GPIO_NUM_13, bump, NULL);

  gpio_set_intr_type(GPIO_NUM_12, GPIO_INTR_POSEDGE); //INDEX_PIN
  gpio_isr_handler_add(GPIO_NUM_12, step, NULL);
}

void setup_camera()
{
    // OV2640 camera module
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;

  if (psramFound()) {
    config.frame_size = FRAMESIZE_UXGA;
    config.jpeg_quality = 10;
    config.fb_count = 2;
  } else {
    config.frame_size = FRAMESIZE_SVGA;
    config.jpeg_quality = 12;
    config.fb_count = 1;
  }
  // Camera init
  esp_err_t err = esp_camera_init(&config);

  if (err != ESP_OK) {
    Serial.printf("Camera init failed with error 0x%x", err);
    // ESP.restart();
    hascamera = false;
  }
  else
    hascamera = true;
    
}

void setup() {

  // touchAttachInterrupt(TOUCH_PIN, gotTouch1, 40);
  pinMode(ENABLE_PIN ,OUTPUT);
  digitalWrite(ENABLE_PIN ,HIGH);

  Serial.begin(115200);
  delay(500);
  Serial.println(F("Serial Initialized, probably."));
  setup_camera();
  Serial.println(F("camera setup."));
  Mount_Storage();
  Serial.println(F("Storage mounted."));
   if (hascamera) capturePhotoSaveSpiffs();
  setup_TMC2209();
  Serial.println(F("Motor setup."));
  connectToWiFi();
  Serial.println(F("WIFI setup."));
  setup_routing(); 
  Serial.println(F("Routing setup."));
  xTaskCreatePinnedToCore(MoveDoor, "Motor", 5000,NULL,5,&Motor,1);
}

void loop() {
    if (hascamera && takepicture)
    {
      takepicture = false;
      capturePhotoSaveSpiffs();
        
    }
    vTaskDelay(1000);
 }

unsigned long move(int start, int spd, bool dir, unsigned long duration)
{
  int i, s, s2, bc, l;
  unsigned long tea, t, t2, threshold, m, currentspd, lastspd, acc, samples;
  long load, lastload;
  
  float a;

  driver.shaft(dir);

  bumped = false;

  tea = millis();
  t = tea;

  l  = duration / 200 ;
  i = 1;
  currentspd = 0;
  m = t;

  while (!bumped) {

      a = abs(sin((M_PI ) * i / l));
      currentspd = start + int(spd * a);

      Serial.printf("Speed %f %d %d (%d) steps %d\n", a, int(spd * a) ,currentspd, driver.cs2rms(driver.cs_actual()), steps);
      driver.VACTUAL(currentspd); //SET SPEED OF MOTOR

      t = m;

      while ((m - t) < 200)
      {
        m = millis();
        delay(10);
        if (bumped) break;
      }

      if (abs(m - tea) > duration) break;
    i++;
  }
  driver.VACTUAL(0);
    
  if (bumped) {
    Serial.printf("bump %d\n", driver.SG_RESULT());
    digitalWrite(RED_LED, LOW);
    delay(200);
  }
  else
    Serial.printf("timed out\n");

  return millis() - tea;
}


void create_json(char *tag, float value) {  
  jsonDocument.clear();
  jsonDocument["type"] = tag;
  jsonDocument["value"] = value;
  serializeJson(jsonDocument, buffer);  
}
 
void add_json_object(char *tag, float value) {
  JsonObject obj = jsonDocument.createNestedObject();
  obj["type"] = tag;
  obj["value"] = value;
}

void OpenDoor(AsyncWebServerRequest * request)
{
  steps = 0;
  direction = -1;
    jsonDocument.clear();
  add_json_object("Bumped", bumped);
  add_json_object("Steps", steps);
  serializeJson(jsonDocument, buffer); 
  request->send(200, "application/json", buffer);
}

void CloseDoor(AsyncWebServerRequest * request)
{
  steps = 0;
  direction = 1;
    jsonDocument.clear();
  add_json_object("Bumped", bumped);
  add_json_object("Steps", steps);
  serializeJson(jsonDocument, buffer); 
  request->send(200, "application/json", buffer);
}
void TakePic(AsyncWebServerRequest * request){
  takepicture = true;
  jsonDocument.clear();
  add_json_object("Bumped", bumped);
  add_json_object("Steps", steps);
  serializeJson(jsonDocument, buffer); 
  request->send(200, "application/json", buffer);  
}

void MoveDoor(void *ptr) {
  unsigned long i;

  while(true) {
    if (direction == 0)
    {
      driver.VACTUAL(0);
      vTaskDelay(1000);
      // Serial.println(".");
      digitalWrite(ENABLE_PIN ,HIGH);
    }
    else
    {
      digitalWrite(ENABLE_PIN ,LOW);
      Serial.println("-");
      shaft =  (direction == 1) ? 1 : 0; 
      digitalWrite(RED_LED, HIGH);
    
      Serial.printf("Moving\n");

      i = move(12800, shaft ? 128 * 1500: 128 * 1500, shaft,shaft ? closetiming : opentiming);
      Serial.printf("Took %d secs.\n", i);
      
      digitalWrite(RED_LED, HIGH);
      vTaskDelay(1000);

      if (bumped) {
        Serial.printf("backing up\n");
        i = move(12800, 0, !shaft, 2000);
        vTaskDelay(1000);
        digitalWrite(RED_LED, HIGH);
      }
      Serial.printf("forward again\n");
      i = move(12800,  0, shaft, 8000);
      direction = 0;
      digitalWrite(RED_LED, HIGH);
    }

  }
}

bool checkPhoto( fs::FS &fs ) {
  File f_pic = fs.open( FILE_PHOTO );
  unsigned int pic_sz = f_pic.size();
  return ( pic_sz > 100 );
}

// Capture Photo and Save it to SPIFFS
void capturePhotoSaveSpiffs( void ) {
  camera_fb_t * fb = NULL; // pointer
  bool ok = 0; // Boolean indicating if the picture has been taken correctly

  do {
    // Take a photo with the camera
    Serial.println("Taking a photo...");

    fb = esp_camera_fb_get();
    if (!fb) {
      Serial.println("Camera capture failed");
      return;
    }

    // Photo file name
    Serial.printf("Picture file name: %s\n", FILE_PHOTO);
    SPIFFS.remove(FILE_PHOTO);

    File file = SPIFFS.open(FILE_PHOTO, FILE_WRITE);

    // Insert the data in the photo file
    if (!file) {
      Serial.println("Failed to open file in writing mode");
    }
    else {
      file.write(fb->buf, fb->len); // payload (image), payload length
      Serial.print("The picture has been saved in ");
      Serial.print(FILE_PHOTO);
      Serial.print(" - Size: ");
      Serial.print(file.size());
      Serial.println(" bytes");
    }
    // Close the file
    file.close();
    esp_camera_fb_return(fb);

    // check if file has been correctly saved in SPIFFS
    ok = checkPhoto(SPIFFS);
  } while ( !ok );
}


void handleUpdate(AsyncWebServerRequest *request) {
  char* html = "<form method='POST' action='/doUpdate' enctype='multipart/form-data'><input type='file' name='update'><input type='submit' value='Update'></form>";
  request->send(200, "text/html", html);
}

void handleDoUpdate(AsyncWebServerRequest *request, const String& filename, size_t index, uint8_t *data, size_t len, bool final) {
  if (!index){
    Serial.println("Update");
    content_len = request->contentLength();
    // if filename includes spiffs, update the spiffs partition
    int cmd = U_FLASH;
    if (!Update.begin(UPDATE_SIZE_UNKNOWN, cmd)) {
      Update.printError(Serial);
    }
  }

  if (Update.write(data, len) != len) {
    Update.printError(Serial);
  }

  if (final) {
    AsyncWebServerResponse *response = request->beginResponse(302, "text/plain", "Please wait while the device reboots");
    response->addHeader("Refresh", "20");  
    response->addHeader("Location", "/");
    request->send(response);
    if (!Update.end(true)){
      Update.printError(Serial);
    } else {
      Serial.println("Update complete");
      Serial.flush();
      ESP.restart();
    }
  }
}

void printProgress(size_t prg, size_t sz) {
  Serial.printf("Progress: %d%%\n", (prg*100)/content_len);
}

